/* Code adapted from pilot-link package */
#include <config.h>
#include <glib.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>
#include <utime.h>
#include <pi-source.h>
#include <pi-socket.h>
#include <pi-file.h>
#include <pi-dlp.h>
#include <stdarg.h>

#include <gpilot-fileman.h>

static void send_error(char *format, ...)
{
  va_list ap;
  char s[1024];
  int len;

  va_start(ap,format);
  vsnprintf(s,sizeof(s),format,ap);
  len=strlen(s);
  write(2,&len,sizeof(len));
  write(2,s,len);
  va_end(ap);
}

static void VoidSyncFlags(int sd)
{
  struct PilotUser U;
  if (dlp_ReadUserInfo(sd, &U)>=0) {
    U.lastSyncPC = 0x00000000; /* Hopefully unique constant, to tell
                                  any Desktop software that databases
                                  have been altered, and that a slow
                                  sync is necessary */
    U.lastSyncDate = U.successfulSyncDate = time(0);
    dlp_WriteUserInfo(sd, &U);
  } 
}

/* FIXME:  Need a better way to disconnect on a signal */
static int global_sd;

static RETSIGTYPE SigHandler(int signal)
{
  if(global_sd)
    pilot_disconnect(global_sd);
  _exit(1);
}

void install_signal_handlers()
{
  signal(SIGHUP, SigHandler);
  signal(SIGINT, SigHandler);
  signal(SIGTERM, SigHandler);
  signal(SIGQUIT, SigHandler);
  signal(SIGSEGV, SigHandler);
}

void pilot_disconnect(int sd)
{
  dlp_EndOfSync(sd, 0);
  pi_close(sd);
  global_sd=0;
}

int pilot_connect(gchar *device) {
  struct pi_sockaddr addr;
  int sd;
  int ret;
  
  if (!(sd = pi_socket(PI_AF_SLP, PI_SOCK_STREAM, PI_PF_PADP))) {
    send_error("pi_socket: %s",strerror(errno));
    return -1;
  }
  
  addr.pi_family = PI_AF_SLP;
  strcpy(addr.pi_device,device);
  
  ret = pi_bind(sd, (struct sockaddr*)&addr, sizeof(addr));
  if(ret == -1) {
    send_error(_("Unable to bind to pilot\n"));
    return -1;
  }
  
  ret = pi_listen(sd,1);
  if(ret == -1) {
    send_error("pi_listen: %s", strerror(errno));
    return -1;
  }

  sd = pi_accept(sd,0,0);
  if(sd == -1) {
    send_error("pi_accept: %s", strerror(errno));
    return -1;
  }

  global_sd=sd;
  return sd;
}

int pilot_install(int sd,char *filename)
{
  struct pi_file * f;

  if ( dlp_OpenConduit(sd) < 0) {
    send_error(_("Exiting on cancel"));
    return -1;
  }

  f = pi_file_open(filename);
  if (f==0) {
    send_error(_("Unable to open '%s'!"), filename);
    return -1;
  }
  fflush(stdout);
  if(pi_file_install(f, sd, 0)<0) {
    send_error(_("Install failed %s."),filename);
    return -1;
  }
  pi_file_close(f);
  
  VoidSyncFlags(sd);

  return 0;
}

static void RemoveFromList(char *name, char **list, int max)
{
  int i;

  for (i = 0; i < max; i++) {
    if (list[i] != NULL && strcmp(name, list[i]) == 0) {
      free(list[i]);
      list[i] = NULL;
    }
  }
}

static void protect_name(char *d, char *s)
{
    while(*s) {
      switch(*s) {
          case '/': *(d++) = '='; *(d++) = '2'; *(d++) = 'F'; break;
          case '=': *(d++) = '='; *(d++) = '3'; *(d++) = 'D'; break;
          case '\x0A': *(d++) = '='; *(d++) = '0'; *(d++) = 'A'; break;
          case '\x0D': *(d++) = '='; *(d++) = '0'; *(d++) = 'D'; break;
       /* If you feel the need:
          case ' ': *(d++) = '='; *(d++) = '2'; *(d++) = '0'; break;
        */
          default: *(d++) = *s;
      }
      ++s;
    }
    *d = '\0';
}

int pilot_backup(int sd,int report_fd,
		 char * dirname, int only_changed, int remove_deleted)
{
  int i, ofile_total, ofile_len;
  DIR * dir;
  struct dirent * dirent;
  char ** orig_files = 0;
  int status;
  
  mkdir(dirname,0700);

  /* Read original list of files in the backup dir */
  ofile_total = 0;
  ofile_len = 0;
  
  if (only_changed) {
    dir = opendir(dirname);
    while( (dirent = readdir(dir)) ) {
      char name[256];
      if (dirent->d_name[0] == '.')
         continue;
         
      if (!orig_files) {
        ofile_len += 256;
      	orig_files = malloc(sizeof(char*) * ofile_len);
      } else if (ofile_total >= ofile_len) {
        ofile_len += 256;
      	orig_files = realloc(orig_files, sizeof(char*) * ofile_len);
      }

      sprintf(name, "%s/%s", dirname, dirent->d_name);
      orig_files[ofile_total++] = strdup(name);
    }
    closedir(dir);
  }

  i = 0;
  for(;;) {
  	struct DBInfo info;
  	struct pi_file * f;
	struct utimbuf times;
	struct stat statb;
        int x;
	//            int skip = 0;
  	char name[256];

        if (dlp_OpenConduit(sd)<0) {
	  send_error(_("Exiting on cancel, all data NOT backed up."));
          return -1;
        }
        
  	if( dlp_ReadDBList(sd, 0, 0x80, i, &info) < 0)
  		break;
  	i = info.index + 1;

	strcpy(name, dirname);
	strcat(name, "/");
	protect_name(name + strlen(name), info.name);

  	if (info.flags & dlpDBFlagResource)
  	  strcat(name,".prc");
  	else
  	  strcat(name,".pdb");

	status=0;
	write(report_fd,&status,sizeof(status));
	status=strlen(name);
	write(2,&status,sizeof(status));
	write(2,name,status);
	
	if (stat(name, &statb) == 0) {
	  if (info.modifyDate == statb.st_mtime) {
	    RemoveFromList(name, orig_files, ofile_total);
	    continue;
	  }
	}

  	/* Ensure that DB-open flag is not kept */
  	info.flags &= 0xff;
  	
  	f = pi_file_create(name, &info);
  	if (f==0) {
  	  send_error(_("Failed, unable to create file\n"));
	  status = -1;
	  write(report_fd,&status,sizeof(status));
	  goto failed_without_open;
	}
  	
  	if(pi_file_retrieve(f, sd, 0)<0) {
	  status = -1;
  	  send_error(_("Failed, unable to back up database\n"));
	  write(report_fd,&status,sizeof(status));
	}
  	pi_file_close(f);
  	
  	/* Note: This is no guarantee that the times on the host system
  	   actually match the GMT times on the Pilot. We only check to
  	   see whether they are the same or different, and do not treat
  	   them as real times. */
  failed_without_open:  	   
	times.actime = info.createDate;
	times.modtime = info.modifyDate;
	utime(name, &times);
	
	RemoveFromList(name, orig_files, ofile_total);
  }
  
  if (orig_files) {
    for (i = 0; i < ofile_total; i++)
      if (orig_files[i] != NULL) {
        if (remove_deleted) {
          unlink(orig_files[i]);
        }
        free(orig_files[i]);
      }
    if (orig_files)
      free(orig_files);
  }

  return 0;
}
